/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/01/10 15:15:32 $
 * $Name: rel0-97 $
 *****************************************************************************/
package mindbright.application;

import java.io.*;

import java.awt.*;
import java.awt.event.*;

import mindbright.ssh.*;
import mindbright.security.*;
import mindbright.terminal.*;

public class MindTerm {

  Frame       frame;
  MenuBar     menubar;
  TerminalWin term;

  SSHInteractiveClient client;
  SSHStdIO             console;
  
  public MindTerm() {
  }

  public static void main(String[] argv) {
    MindTerm controller = new MindTerm();
    controller.startup(argv);
  }

  public void startup(String[] argv) {
    try {
      String  server;
      String  commandLine    = null;
      String  username       = null;
      String  sshHome        = null;
      String  idFile         = null;
      String  cipher         = null;
      String  authType       = null;
      boolean haveGUI        = true;
      int     port           = -1;
      int     i;

      client  = new SSHInteractiveClient();
      console = (SSHStdIO)client.getConsole();

      for(i = 0; i < argv.length; i++) {
	String arg = argv[i];
	if(arg.charAt(0) != '-')
	  break;
	switch(arg.charAt(1)) {
	case 'l':
	  username = argv[++i];
	  break;
	case 'i':
	  idFile = argv[++i];
	  break;
	case 'd':
	  haveGUI = false;
	  break;
	case 'v':
	  System.out.println("verbose mode selected...");
	  SSH.DEBUG = true;
	  break;
	case 'V':
	  System.out.println("SSH version " + SSH.VER_SSH + ", protocol version " +
			     SSH.VER_MAJOR + "." + SSH.VER_MINOR);
	  System.exit(0);
	  break;
	case 'c':
	  cipher = argv[++i];
	  break;
	case 'a':
	  authType = argv[++i];
	  break;
	case 'p':
	  port = Integer.parseInt(argv[++i]);
	  break;
	case 'L':
	  client.addLocalPortForward(argv[++i]);
	  break;
	case 'R':
	  client.addRemotePortForward(argv[++i]);
	  break;
	default:
	  throw new Exception();
	}
      }

      if(i < argv.length) {
	server = argv[i++];
	client.setSrvHost(server);
      }

      if(i < argv.length) {
	commandLine = "";
	for(; i < argv.length; i++) {
	  commandLine += argv[i] + " ";
	}
	haveGUI = false;
      }

      if(port > 0)
	client.setSrvPort(port);

      // Fetch username on local system if empty
      //
      if(username == null)
	username = System.getProperty("user.name");
      if(username != null)
	client.setUsername(username);

      sshHome = System.getProperty("user.home") + File.separator + ".ssh" + File.separator;
      if(idFile == null) {
	idFile   = "identity";
      } else {
	i = idFile.lastIndexOf(File.separatorChar);
	if(i != -1)
	  sshHome = idFile.substring(0, i + 1);
	idFile   = idFile.substring(i + 1);
      }
      client.setSSHHome(sshHome);
      client.setIdentityFileName(idFile);

      if(cipher != null)
	client.setCipher(cipher);
      if(authType != null)
	client.setAuthType(authType);

      Thread clientThread = new Thread(client);
      if(haveGUI) {
	initGUI();
	console.setTerminal(term);
	console.setOwnerContainer(frame);
	console.setOwnerName("MindTerm v0.97");
	console.updateTitle();
      }

      if(commandLine != null) {
	try {
	  client.doSingleCommand(commandLine, false, 0);
	} catch (IOException e) {
	  System.out.println("error: "  + e.getMessage());
	}
      } else {
	// Adhoc... Let the AWT-stuff settle before we try to print to the term...
	// (TODO: get rid of this, should wait for something more specific I guess :-)
	//
	Thread.sleep(1000);
	clientThread.start();
	clientThread.join();
	windowClosing(null);
      }

    } catch(Exception e) {
      System.out.println("usage: MindTerm [options] [<server> [<command>]]");
      System.out.println("Options:");
      System.out.println("  -l user     Log in using this user name.");
      System.out.println("  -d          No terminal-window, only dumb command-line and port-forwarding.");
      System.out.println("  -i file     Name of the RSA identity file (default: ~/.ssh/identity).");
      System.out.println("  -v          Verbose; display verbose debugging messages.");
      System.out.println("  -V          Display version number only.");
      System.out.println("  -a authtype Select authentication type: ``passwd'', ``rsa'', ``rhostsrsa''");
      System.out.println("  -c cipher   Select encryption algorithm: ``blowfish'', ``3des'', ``des''...");
      System.out.println("  -p port     Connect to this port.  Server must be on the same port.");
      System.out.println("  -L [/plugin/]listen-port:host:port Forward local port to remote address.");
      System.out.println("  -R [/plugin/]listen-port:host:port Forward remote port to local address.");

      windowClosing(null);
    }
  }

  public void initGUI() {
    term = new TerminalWin(80, 24, new TerminalXTerm());

    // !!!
    term.setSaveLines(1024);

    frame = new Frame();
    frame.setLayout(new BorderLayout());
    frame.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e)  { MindTerm.this.windowClosing(e); }

      // !!! I don't get it, why do this occur? Who gets focus when we
      // are deiconified??  (this do only apply to winDOS)
      // NOTE:
      // Without this AND the FocusListener.focusLost() in
      // TerminalWin.java we don't regain keyboard-focus after having
      // been iconified !!!
      //
      public void windowDeiconified(WindowEvent e)  { term.requestFocus(); }
    });

    // !!! Weird, on Linux if the menubar is added here before the window is visible the
    // TerminalWin won't get it's events... TODO: check this
    //
    // frame.setMenuBar(MenuHandler.createMenus());

    frame.add(term, "Center");
    frame.add(term.createScrollbar(), "East");

    frame.pack();
    frame.show();
  }

  public void windowClosing(WindowEvent e) {
    System.out.println("Thank you for using MindTerm...");
    if(frame != null)
      frame.dispose();
    System.exit(0);
  }

}

