/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/01/10 15:15:37 $
 * $Name: rel0-97 $
 *****************************************************************************/
package mindbright.ssh;

import java.net.*;
import java.io.*;

public class SSHListenChannel extends SSHChannel {
  SSHChannelController controller;

  ServerSocket listenSocket;
  String       remoteHost;
  int          remotePort;

  boolean temporaryListener;

  public SSHListenChannel(int localPort, String remoteHost, int remotePort,
			  SSHChannelController controller)
    throws IOException {
    super(SSH.LISTEN_CHAN_NUM);
    this.controller    = controller;
    this.listenSocket  = new ServerSocket(localPort);
    this.remoteHost    = remoteHost;
    this.remotePort    = remotePort;
  }

  public int getListenPort() {
    return listenSocket.getLocalPort();
  }

  public SSHTunnel newTunnel(Socket ioSocket, int channelId, int remoteChannelId,
			     SSHChannelController controller) throws IOException {
    return new SSHTunnel(ioSocket, channelId, remoteChannelId, controller);
  }

  public void setTemporaryListener(boolean val) {
    temporaryListener = val;
  }

  public void serviceLoop() throws IOException {

    SSH.log("Starting listen-chan: " + listenSocket.getLocalPort());

    for(;;) {
      Socket fwdSocket = listenSocket.accept();
      SSHPduOutputStream respPdu = new SSHPduOutputStream(SSH.MSG_PORT_OPEN, controller.sndCipher);

      int newChan      = controller.newChannelId();
      SSHTunnel tunnel = newTunnel(fwdSocket,
				   newChan, SSH.UNKNOWN_CHAN_NUM,
				   controller);
      controller.addTunnel(tunnel);
      tunnel.setRemoteDesc(remoteHost + ":" + remotePort);

      respPdu.writeInt(newChan);
      respPdu.writeString(remoteHost);
      respPdu.writeInt(remotePort);

      SSH.log("got connect for: " + remoteHost + " : " + remotePort + ", " + newChan);

      //
      // !!! TODO: check this !!! if(controller.haveHostInFwdOpen())
      //
      respPdu.writeString(fwdSocket.getInetAddress().getHostAddress());

      controller.transmit(respPdu);

      if(temporaryListener)
	break;
    }

    listenSocket.close();
  }

}
