/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/01/10 15:15:37 $
 * $Name: rel0-97 $
 *****************************************************************************/
package mindbright.ssh;

import java.util.Vector;

public class SSHPduQueue {

  Vector     queue;
  boolean    isWaiting;
  int        maxDepth;
  int        curDepth;

  public SSHPduQueue() {
    this.queue     = new Vector();
    this.isWaiting = false;
    this.maxDepth  = 65535;
    this.curDepth  = 0;
  }

  public void setMaxDepth(int maxDepth) {
    synchronized(queue) {
      this.maxDepth = maxDepth;
    }
  }

  public void addLast(SSHPdu pdu) {
    synchronized(queue) {
      if(curDepth++ < maxDepth)
	queue.addElement(pdu);
      if(isWaiting)
	queue.notify();
    }
  }

  public void addFirst(SSHPdu pdu) {
    synchronized(queue) {
      if(curDepth++ < maxDepth)
	queue.insertElementAt(pdu, 0);
      if(isWaiting)
	queue.notify();
    }
  }

  public void release() {
    synchronized(queue) {
      if(isWaiting)
	queue.notify();
    }
  }

  public boolean isEmpty() {
    boolean isEmpty;
    synchronized(queue) {
      isEmpty = queue.isEmpty();
    }
    return isEmpty;
  }

  public SSHPdu getFirst() {
    boolean doSleep = false;
    SSHPdu  pdu     = null;

    synchronized(queue) {
      if(queue.isEmpty()) {
	isWaiting = true;
	try {
	  queue.wait();
	} catch (InterruptedException e) {
	  // !!!
	}
      }
      isWaiting = false;
      try {
	pdu = (SSHPdu) queue.firstElement();
	queue.removeElementAt(0);
	curDepth--;
      } catch (Exception e) {
	pdu = null;
      }
    }

    return pdu;
  }

}
