/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/01/10 15:15:38 $
 * $Name: rel0-97 $
 *****************************************************************************/
package mindbright.ssh;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import java.math.BigInteger;

import mindbright.security.*;

public class SSHRSAPublicKeyFile {

  Vector pubKeyList;

  public SSHRSAPublicKeyFile(String name) throws IOException {
    FileInputStream fileIn = new FileInputStream(name);
    BufferedReader  reader = new BufferedReader(new InputStreamReader(fileIn));
    String          row;

    pubKeyList = new Vector();

    while((row = reader.readLine()) != null) {
      row = row.trim();
      if(row.equals("") || row.charAt(0) == '#') // Skip comment-lines and empty lines...
	continue;
      String opts = ""; // !!! Read options from start of line...
      try {
	SSHRSAPublicKeyString pubKey = SSHRSAPublicKeyString.createKey(opts, row);
	pubKeyList.addElement(pubKey);
      } catch (Exception e) {
	throw new IOException("Corrupt public keys file: " + name);
      }
    }

  }

  public Enumeration elements() {
    return pubKeyList.elements();
  }

  public RSAPublicKey getPublic(BigInteger n, String user) {
    SSHRSAPublicKeyString pk = null;

    Enumeration e = pubKeyList.elements();
    while(e.hasMoreElements()) {
      pk = (SSHRSAPublicKeyString) e.nextElement();
      if(pk.getN().equals(n))
	break;
      pk = null;
    }

    return pk;
  }

  public static void main(String[] argv) {
    SSHRSAPublicKeyFile file = null;

    try {
      file = new SSHRSAPublicKeyFile("/home/mats/.ssh/authorized_keys");

      SSHRSAPublicKeyString pk = null;

      Enumeration e = file.elements();
      while(e.hasMoreElements()) {
	pk = (SSHRSAPublicKeyString) e.nextElement();
	System.out.println(pk);
      }
    } catch (Exception e) {
      e.printStackTrace();
      System.out.println("Error: " + e.toString());
    }
  }


}


